import os
from collections import OrderedDict
from contextlib import contextmanager

import pytest
from click.testing import CliRunner
from dagster_celery import celery_executor
from dagster_celery.cli import main

from dagster import (
    ExecutionTargetHandle,
    ModeDefinition,
    default_executors,
    execute_pipeline,
    pipeline,
    seven,
    solid,
)
from dagster.core.instance import DagsterInstance

BUILDKITE = os.getenv('BUILDKITE')
skip_ci = pytest.mark.skipif(
    bool(BUILDKITE),
    reason='Tests hang forever on buildkite for reasons we don\'t currently understand',
)

celery_mode_defs = [ModeDefinition(executor_defs=default_executors + [celery_executor])]


@contextmanager
def execute_pipeline_on_celery(pipeline_name):
    with seven.TemporaryDirectory() as tempdir:
        handle = ExecutionTargetHandle.for_pipeline_python_file(__file__, pipeline_name)
        pipeline_def = handle.build_pipeline_definition()
        instance = DagsterInstance.local_temp(tempdir=tempdir)
        result = execute_pipeline(
            pipeline_def,
            environment_dict={
                'storage': {'filesystem': {'config': {'base_dir': tempdir}}},
                'execution': {'celery': {}},
            },
            instance=instance,
        )
        yield result


@contextmanager
def execute_eagerly_on_celery(pipeline_name):
    with seven.TemporaryDirectory() as tempdir:
        result = execute_pipeline(
            ExecutionTargetHandle.for_pipeline_python_file(
                __file__, pipeline_name
            ).build_pipeline_definition(),
            environment_dict={
                'storage': {'filesystem': {'config': {'base_dir': tempdir}}},
                'execution': {'celery': {'config': {'config_source': {'task_always_eager': True}}}},
            },
            instance=DagsterInstance.local_temp(tempdir=tempdir),
        )
        yield result


@contextmanager
def start_celery_worker():
    runner = CliRunner()
    result = runner.invoke(main, ['worker', 'start', '-d'])
    assert result.exit_code == 0, str(result.exception)

    yield

    result = runner.invoke(main, ['worker', 'terminate'])
    assert result.exit_code == 0, str(result.exception)


@solid(tags={'dagster-celery/priority': 0})
def zero(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '0'
    context.log.info('Executing with priority 0')
    return True


@solid(tags={'dagster-celery/priority': 1})
def one(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '1'
    context.log.info('Executing with priority 1')
    return True


@solid(tags={'dagster-celery/priority': 2})
def two(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '2'
    context.log.info('Executing with priority 2')
    return True


@solid(tags={'dagster-celery/priority': 3})
def three(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '3'
    context.log.info('Executing with priority 3')
    return True


@solid(tags={'dagster-celery/priority': 4})
def four(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '4'
    context.log.info('Executing with priority 4')
    return True


@solid(tags={'dagster-celery/priority': 5})
def five(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '5'
    context.log.info('Executing with priority 5')
    return True


@solid(tags={'dagster-celery/priority': 6})
def six(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '6'
    context.log.info('Executing with priority 6')
    return True


@solid(tags={'dagster-celery/priority': 7})
def seven_(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '7'
    context.log.info('Executing with priority 7')
    return True


@solid(tags={'dagster-celery/priority': 8})
def eight(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '8'
    context.log.info('Executing with priority 8')
    return True


@solid(tags={'dagster-celery/priority': 9})
def nine(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '9'
    context.log.info('Executing with priority 9')
    return True


@solid(tags={'dagster-celery/priority': 10})
def ten(context):
    assert 'dagster-celery/priority' in context.solid.tags
    assert context.solid.tags['dagster-celery/priority'] == '10'
    context.log.info('Executing with priority 10')
    return True


@pipeline(mode_defs=celery_mode_defs)
def priority_pipeline():
    for i in range(50):
        zero.alias('zero_' + str(i))()
        one.alias('one_' + str(i))()
        two.alias('two_' + str(i))()
        three.alias('three_' + str(i))()
        four.alias('four_' + str(i))()
        five.alias('five_' + str(i))()
        six.alias('six_' + str(i))()
        seven_.alias('seven_' + str(i))()
        eight.alias('eight_' + str(i))()
        nine.alias('nine' + str(i))()
        ten.alias('ten_' + str(i))()


@pytest.mark.skip
@skip_ci
def test_priority_pipeline():
    with execute_pipeline_on_celery('priority_pipeline') as result:
        assert result.success


@pipeline(mode_defs=celery_mode_defs)
def simple_priority_pipeline():
    zero()
    one()
    two()
    three()
    four()
    five()
    six()
    seven_()
    eight()
    nine()
    ten()


def test_eager_priority_pipeline():
    with execute_eagerly_on_celery('simple_priority_pipeline') as result:
        assert result.success
        assert list(OrderedDict.fromkeys([evt.step_key for evt in result.step_event_list])) == [
            'ten.compute',
            'nine.compute',
            'eight.compute',
            'seven_.compute',
            'six.compute',
            'five.compute',
            'four.compute',
            'three.compute',
            'two.compute',
            'one.compute',
            'zero.compute',
        ]



from celery import Celery
from celery.utils.collections import force_mapping
from dagster_celery.config import CeleryConfig
from dagster_graphql.client.mutations import execute_execute_plan_mutation
from kombu import Queue

from dagster import ExecutionTargetHandle, check
from dagster.core.instance import InstanceRef
from dagster.core.serdes import serialize_dagster_namedtuple
from dagster.seven import is_module_available


def create_task(celery_app, **task_kwargs):
    @celery_app.task(bind=True, name='execute_query', **task_kwargs)
    def _execute_query(_self, handle_dict, variables, instance_ref_dict):
        instance_ref = InstanceRef.from_dict(instance_ref_dict)
        handle = ExecutionTargetHandle.from_dict(handle_dict)

        events = execute_execute_plan_mutation(
            handle=handle, variables=variables, instance_ref=instance_ref,
        )
        serialized_events = [serialize_dagster_namedtuple(event) for event in events]
        return serialized_events

    return _execute_query


def make_app(config=None):
    config = check.opt_inst_param(config, 'config', CeleryConfig)
    app_ = Celery('dagster', **(config._asdict() if config is not None else {}))
    if config is None:
        app_.config_from_object('dagster_celery.defaults', force=True)

        if is_module_available('dagster_celery_config'):
            # pylint: disable=protected-access
            obj = force_mapping(app_.loader._smart_import('dagster_celery_config'))
            app_.conf.update(obj)

    app_.loader.import_module('celery.contrib.testing.tasks')

    app_.conf.task_queues = [
        Queue('dagster', routing_key='dagster.#', queue_arguments={'x-max-priority': 10})
    ]
    app_.conf.task_routes = {
        'execute_query': {'queue': 'dagster', 'routing_key': 'dagster.execute_query'}
    }
    app_.conf.task_queue_max_priority = 10
    app_.conf.task_default_priority = 5
    return app_


app = make_app()

execute_query = create_task(app)



import pytest
from celery.contrib.testing import worker
from celery.contrib.testing.app import setup_default_app
from dagster_celery.tasks import make_app


@pytest.fixture(scope='session')
def dagster_celery_app():
    app = make_app()
    with setup_default_app(app, use_trap=False):
        yield app


# pylint doesn't understand pytest fixtures
@pytest.fixture(scope='function')
def dagster_celery_worker(dagster_celery_app):  # pylint: disable=redefined-outer-name
    with worker.start_worker(dagster_celery_app, perform_ping_check=False) as w:
        yield w



import os
import sys
import time
from contextlib import contextmanager

import pytest
from click.testing import CliRunner
from dagster_celery.cli import main

from dagster import check
from dagster.seven import mock
from dagster.utils import file_relative_path

BUILDKITE = os.getenv('BUILDKITE')

skip_ci = pytest.mark.skipif(
    bool(BUILDKITE),
    reason='Tests hang forever on buildkite for reasons we don\'t currently understand',
)


def assert_called(mck):
    if hasattr(mck, 'assert_called'):
        mck.assert_called()
    else:
        # py35
        assert mck.called


@contextmanager
def pythonpath(path):
    '''Inserts a path into the PYTHONPATH, then restores the previous PYTHONPATH when it
    cleans up'''
    unset = 'PYTHONPATH' not in os.environ
    old_path = os.environ.get('PYTHONPATH', '')
    old_sys_path = sys.path
    try:
        os.environ['PYTHONPATH'] = '{old_path}{path}:'.format(old_path=old_path, path=path)
        sys.path.insert(0, path)
        yield
    finally:
        if unset:
            del os.environ['PYTHONPATH']
        else:
            os.environ['PYTHONPATH'] = old_path
        sys.path = old_sys_path


def start_worker(name, args=None, exit_code=0, exception_str=''):
    args = check.opt_list_param(args, 'args')
    runner = CliRunner()
    result = runner.invoke(main, ['worker', 'start', '-d'] + args + ['--name', name])
    assert result.exit_code == exit_code, str(result.exception)
    if exception_str:
        assert exception_str in str(result.exception)


@contextmanager
def cleanup_worker(name, args=None):
    args = check.opt_list_param(args, 'args')
    try:
        yield
    finally:
        runner = CliRunner()
        result = runner.invoke(main, ['worker', 'terminate'] + args + [name])
        assert result.exit_code == 0, str(result.exception)


def check_for_worker(name, args=None, present=True):
    runner = CliRunner()
    args = check.opt_list_param(args, 'args')
    result = runner.invoke(main, ['worker', 'list'] + args)
    assert result.exit_code == 0, str(result.exception)
    retry_count = 0
    while retry_count < 10 and (
        not '{name}@'.format(name=name) in result.output
        if present
        else '{name}@'.format(name=name) in result.output
    ):
        time.sleep(1)
        result = runner.invoke(main, ['worker', 'list'] + args)
        assert result.exit_code == 0, str(result.exception)
        retry_count += 1
    return (
        '{name}@'.format(name=name) in result.output
        if present
        else '{name}@'.format(name=name) not in result.output
    )


def test_invoke_entrypoint():
    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 0
    assert 'worker' in result.output

    runner = CliRunner()
    result = runner.invoke(main, ['worker'])
    assert result.exit_code == 0
    assert 'Start a dagster celery worker' in result.output


@skip_ci
def test_start_worker():
    with cleanup_worker('dagster_test_worker'):
        start_worker('dagster_test_worker')
        assert check_for_worker('dagster_test_worker')


@skip_ci
def test_start_worker_too_many_queues():
    args = ['-q', '1', '-q', '2', '-q', '3', '-q', '4', '-q', '5']

    with cleanup_worker('dagster_test_worker'):
        start_worker(
            'dagster_test_worker',
            args=args,
            exit_code=1,
            exception_str=(
                'Can\'t start a dagster_celery worker that listens on more than four queues, due to a '
                'bug in Celery 4.'
            ),
        )


@skip_ci
def test_start_worker_config_from_empty_yaml():
    args = ['-y', file_relative_path(__file__, 'empty.yaml')]
    with cleanup_worker('dagster_test_worker', args=args):
        start_worker('dagster_test_worker', args=args)
        assert check_for_worker('dagster_test_worker')


@skip_ci
def test_start_worker_config_from_yaml():
    args = ['-y', file_relative_path(__file__, 'engine_config.yaml')]

    with cleanup_worker('dagster_test_worker', args=args):
        start_worker('dagster_test_worker', args=args)
        assert check_for_worker('dagster_test_worker')


@mock.patch('dagster_celery.cli.launch_background_worker')
def test_mock_start_worker(worker_patch):
    start_worker('dagster_test_worker')
    assert_called(worker_patch)


@mock.patch('dagster_celery.cli.launch_background_worker')
def test_mock_start_worker_config_from_empty_yaml(worker_patch):
    args = ['-y', file_relative_path(__file__, 'empty.yaml')]
    start_worker('dagster_test_worker', args=args)
    assert_called(worker_patch)


@mock.patch('dagster_celery.cli.launch_background_worker')
def test_start_mock_worker_config_from_yaml(worker_patch):
    args = ['-y', file_relative_path(__file__, 'engine_config.yaml')]
    start_worker('dagster_test_worker', args=args)
    assert_called(worker_patch)


@mock.patch('dagster_celery.cli.launch_background_worker')
def test_mock_start_worker_too_many_queues(_worker_patch):
    args = ['-q', '1', '-q', '2', '-q', '3', '-q', '4', '-q', '5']

    start_worker(
        'dagster_test_worker',
        args=args,
        exit_code=1,
        exception_str=(
            'Can\'t start a dagster_celery worker that listens on more than four queues, due to a '
            'bug in Celery 4.'
        ),
    )



from collections import namedtuple

from dagster import check
from dagster.core.execution.config import ExecutorConfig

from .defaults import (
    broker_transport_options,
    broker_url,
    result_backend,
    task_default_priority,
    task_default_queue,
)

DEFAULT_CONFIG = {
    # 'task_queue_max_priority': 10,
    'worker_prefetch_multiplier': 1,
    'broker_transport_options': broker_transport_options,
    'task_default_priority': task_default_priority,
    'task_default_queue': task_default_queue,
}


class dict_wrapper(object):
    '''Wraps a dict to convert `obj['attr']` to `obj.attr`.'''

    def __init__(self, dictionary):
        self.__dict__ = dictionary


class CeleryConfig(
    namedtuple('CeleryConfig', 'broker backend include config_source'), ExecutorConfig,
):
    '''Configuration class for the Celery execution engine.

    Params:
        broker (Optional[str]): The URL of the Celery broker.
        backend (Optional[str]): The URL of the Celery backend.
        include (Optional[List[str]]): List of modules every worker should import.
        config_source (Optional[Dict]): Config settings for the Celery app.

    '''

    def __new__(
        cls, broker=None, backend=None, include=None, config_source=None,
    ):

        return super(CeleryConfig, cls).__new__(
            cls,
            broker=check.opt_str_param(broker, 'broker', default=broker_url),
            backend=check.opt_str_param(backend, 'backend', default=result_backend),
            include=check.opt_list_param(include, 'include', of_type=str),
            config_source=dict_wrapper(
                dict(DEFAULT_CONFIG, **check.opt_dict_param(config_source, 'config_source'))
            ),
        )

    @staticmethod
    def get_engine():
        from .engine import CeleryEngine

        return CeleryEngine()



import sys
import time
from collections import defaultdict

from dagster import check
from dagster.core.engine.engine_base import Engine
from dagster.core.errors import DagsterSubprocessError
from dagster.core.events import DagsterEvent, EngineEventData
from dagster.core.execution.context.system import SystemPipelineExecutionContext
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.core.serdes import deserialize_json_to_dagster_namedtuple
from dagster.utils.error import serializable_error_info_from_exc_info
from dagster.utils.net import is_local_uri

from .config import CeleryConfig
from .defaults import task_default_priority, task_default_queue
from .tasks import create_task, make_app

TICK_SECONDS = 1


class CeleryEngine(Engine):
    @staticmethod
    def execute(pipeline_context, execution_plan):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)

        check.param_invariant(
            isinstance(pipeline_context.executor_config, CeleryConfig),
            'pipeline_context',
            'Expected executor_config to be CeleryConfig got {}'.format(
                pipeline_context.executor_config
            ),
        )

        celery_config = pipeline_context.executor_config

        storage = pipeline_context.environment_dict.get('storage')

        if (celery_config.broker and not is_local_uri(celery_config.broker)) or (
            celery_config.backend and not is_local_uri(celery_config.backend)
        ):
            check.invariant(
                storage.get('s3') or storage.get('gcs'),
                'Must use S3 or GCS storage with non-local Celery broker: {broker} '
                'and backend: {backend}'.format(
                    broker=celery_config.broker, backend=celery_config.backend
                ),
            )
        else:
            check.invariant(
                not storage.get('in_memory'),
                'Cannot use in-memory storage with Celery, use filesystem, S3, or GCS',
            )

        pipeline_name = pipeline_context.pipeline_def.name

        handle_dict = pipeline_context.execution_target_handle.to_dict()

        instance_ref_dict = pipeline_context.instance.get_ref().to_dict()

        environment_dict = dict(pipeline_context.environment_dict, execution={'in_process': {}})

        mode = pipeline_context.mode_def.name

        run_id = pipeline_context.pipeline_run.run_id

        app = make_app(celery_config)

        task_signatures = {}  # Dict[step_key, celery.Signature]
        apply_kwargs = defaultdict(dict)  # Dict[step_key, Dict[str, Any]]

        priority_for_step = lambda step: (
            -1 * int(step.tags.get('dagster-celery/priority', task_default_priority))
        )
        priority_for_key = lambda step_key: (-1 * apply_kwargs[step_key]['priority'])
        _warn_on_priority_misuse(pipeline_context, execution_plan)

        for step_key in execution_plan.step_keys_to_execute:
            step = execution_plan.get_step_by_key(step_key)
            priority = int(step.tags.get('dagster-celery/priority', task_default_priority))
            queue = step.tags.get('dagster-celery/queue', task_default_queue)
            task = create_task(app)

            variables = {
                'executionParams': {
                    'selector': {'name': pipeline_name},
                    'environmentConfigData': environment_dict,
                    'mode': mode,
                    'executionMetadata': {'runId': run_id},
                    'stepKeys': [step_key],
                }
            }
            task_signatures[step_key] = task.si(handle_dict, variables, instance_ref_dict)
            apply_kwargs[step_key] = {
                'priority': priority,
                'queue': queue,
                'routing_key': '{queue}.execute_query'.format(queue=queue),
            }

        step_results = {}  # Dict[ExecutionStep, celery.AsyncResult]
        step_errors = {}
        completed_steps = set({})  # Set[step_key]
        active_execution = execution_plan.start(sort_key_fn=priority_for_step)
        stopping = False

        while (not active_execution.is_complete and not stopping) or step_results:

            results_to_pop = []
            for step_key, result in sorted(
                step_results.items(), key=lambda x: priority_for_key(x[0])
            ):
                if result.ready():
                    try:
                        step_events = result.get()
                    except Exception as e:  # pylint: disable=broad-except
                        # We will want to do more to handle the exception here.. maybe subclass Task
                        # Certainly yield an engine or pipeline event
                        step_events = []
                        step_errors[step_key] = serializable_error_info_from_exc_info(
                            sys.exc_info()
                        )
                        stopping = True
                    for step_event in step_events:
                        event = deserialize_json_to_dagster_namedtuple(step_event)
                        yield event
                        active_execution.handle_event(event)

                    results_to_pop.append(step_key)
                    completed_steps.add(step_key)

            for step_key in results_to_pop:
                if step_key in step_results:
                    del step_results[step_key]
                    active_execution.verify_complete(pipeline_context, step_key)

            # process skips from failures or uncovered inputs
            for event in active_execution.skipped_step_events_iterator(pipeline_context):
                yield event

            # dont add any new steps if we are stopping
            if stopping:
                continue

            # This is a slight refinement. If we have n workers idle and schedule m > n steps for
            # execution, the first n steps will be picked up by the idle workers in the order in
            # which they are scheduled (and the following m-n steps will be executed in priority
            # order, provided that it takes longer to execute a step than to schedule it). The test
            # case has m >> n to exhibit this behavior in the absence of this sort step.
            for step in active_execution.get_steps_to_execute():
                try:
                    step_results[step.key] = task_signatures[step.key].apply_async(
                        **apply_kwargs[step.key]
                    )
                except Exception:
                    yield DagsterEvent.engine_event(
                        pipeline_context,
                        'Encountered error during celery task submission.'.format(),
                        event_specific_data=EngineEventData.engine_error(
                            serializable_error_info_from_exc_info(sys.exc_info()),
                        ),
                    )
                    raise

            time.sleep(TICK_SECONDS)

        if step_errors:
            raise DagsterSubprocessError(
                'During celery execution errors occured in workers:\n{error_list}'.format(
                    error_list='\n'.join(
                        [
                            '[{step}]: {err}'.format(step=key, err=err.to_string())
                            for key, err in step_errors.items()
                        ]
                    )
                ),
                subprocess_error_infos=list(step_errors.values()),
            )


def _warn_on_priority_misuse(context, execution_plan):
    bad_keys = []
    for key in execution_plan.step_keys_to_execute:
        step = execution_plan.get_step_by_key(key)
        if (
            step.tags.get('dagster/priority') is not None
            and step.tags.get('dagster-celery/priority') is None
        ):
            bad_keys.append(key)

    if bad_keys:
        context.log.warn(
            'The following steps do not have "dagster-celery/priority" set but do '
            'have "dagster/priority" set which is not applicable for the celery engine: [{}]. '
            'Consider using a function to set both keys.'.format(', '.join(bad_keys))
        )



import os
import threading
import time
from contextlib import contextmanager

import pytest
from click.testing import CliRunner
from dagster_celery import celery_executor
from dagster_celery.cli import main

from dagster import (
    ExecutionTargetHandle,
    ModeDefinition,
    default_executors,
    execute_pipeline,
    pipeline,
    seven,
    solid,
)
from dagster.core.instance import DagsterInstance

BUILDKITE = os.getenv('BUILDKITE')
skip_ci = pytest.mark.skipif(
    bool(BUILDKITE),
    reason='Tests hang forever on buildkite for reasons we don\'t currently understand',
)

celery_mode_defs = [ModeDefinition(executor_defs=default_executors + [celery_executor])]


@contextmanager
def execute_pipeline_on_celery(pipeline_name):
    with seven.TemporaryDirectory() as tempdir:
        handle = ExecutionTargetHandle.for_pipeline_python_file(__file__, pipeline_name)
        pipeline_def = handle.build_pipeline_definition()
        instance = DagsterInstance.local_temp(tempdir=tempdir)
        result = execute_pipeline(
            pipeline_def,
            environment_dict={
                'storage': {'filesystem': {'config': {'base_dir': tempdir}}},
                'execution': {'celery': {}},
            },
            instance=instance,
        )
        yield result


@contextmanager
def start_celery_worker(queue=None):
    runner = CliRunner()
    runargs = ['worker', 'start', '-d']
    if queue:
        runargs += ['-q', queue]
    result = runner.invoke(main, runargs)
    assert result.exit_code == 0, str(result.exception)
    try:
        yield
    finally:
        result = runner.invoke(main, ['worker', 'terminate'])
        assert result.exit_code == 0, str(result.exception)


@solid(tags={'dagster-celery/queue': 'fooqueue'})
def fooqueue(context):
    assert context.solid.tags['dagster-celery/queue'] == 'fooqueue'
    context.log.info('Executing on queue fooqueue')
    return True


@pipeline(mode_defs=celery_mode_defs)
def multiqueue_pipeline():
    fooqueue()


# @pytest.mark.skip
@skip_ci
def test_multiqueue():
    def execute_on_thread(done):
        with execute_pipeline_on_celery('multiqueue_pipeline') as result:
            assert result.success
            done.set()

    done = threading.Event()
    with start_celery_worker():
        execute_thread = threading.Thread(target=execute_on_thread, args=(done,))
        execute_thread.daemon = True
        execute_thread.start()
        time.sleep(1)
        assert not done.is_set()
        with start_celery_worker(queue='fooqueue'):
            execute_thread.join()
            assert done.is_set()



import os
import subprocess
import uuid

import click
from celery.utils.nodenames import default_nodename, host_format

from dagster import check
from dagster.config.validate import validate_config
from dagster.core.instance import DagsterInstance
from dagster.utils import load_yaml_from_path, mkdir_p

from .config import CeleryConfig
from .executor import celery_executor
from .tasks import make_app


def create_worker_cli_group():
    group = click.Group(name='worker')
    group.add_command(worker_start_command)
    group.add_command(worker_terminate_command)
    group.add_command(worker_list_command)
    return group


def get_config_value_from_yaml(yaml_path):
    if yaml_path is None:
        return {}
    parsed_yaml = load_yaml_from_path(yaml_path) or {}
    # Would be better not to hardcode this path
    return parsed_yaml.get('execution', {}).get('celery', {}) or {}


def get_app(config_yaml=None):
    if config_yaml:
        celery_config = CeleryConfig(**get_config_value_from_yaml(config_yaml))
    else:
        celery_config = CeleryConfig()

    app = make_app(celery_config)

    return app


def get_worker_name(name=None):
    return (
        name + '@%h'
        if name is not None
        else 'dagster-{uniq}@%h'.format(uniq=str(uuid.uuid4())[-6:])
    )


def get_config_dir(config_yaml=None):
    instance = DagsterInstance.get()
    config_type = celery_executor.config_field.config_type
    config_value = get_config_value_from_yaml(config_yaml)

    config_module_name = 'dagster_celery_config'

    config_dir = os.path.join(
        instance.root_directory, 'dagster_celery', 'config', str(uuid.uuid4())
    )
    mkdir_p(config_dir)
    config_path = os.path.join(
        config_dir, '{config_module_name}.py'.format(config_module_name=config_module_name)
    )
    validated_config = validate_config(config_type, config_value).value
    with open(config_path, 'w') as fd:
        if 'broker' in validated_config:
            fd.write(
                'broker_url = \'{broker_url}\'\n'.format(broker_url=str(validated_config['broker']))
            )
        if 'backend' in validated_config:
            fd.write(
                'result_backend = \'{result_backend}\'\n'.format(
                    result_backend=str(validated_config['backend'])
                )
            )
        if 'config_source' in validated_config:
            for key, value in validated_config['config_source'].items():
                fd.write('{key} = {value}\n'.format(key=key, value=repr(value)))

    # n.b. right now we don't attempt to clean up this cache, but it might make sense to delete
    # any files older than some time if there are more than some number of files present, etc.
    return config_dir


def launch_background_worker(subprocess_args, env):
    return subprocess.Popen(
        subprocess_args + ['--detach', '--pidfile='], stdout=None, stderr=None, env=env
    )


@click.command(name='start', help='Start a dagster celery worker.')
@click.option(
    '--name',
    '-n',
    type=click.STRING,
    default=None,
    help=(
        'The name of the worker. Defaults to a unique name prefixed with "dagster-" and ending '
        'with the hostname.'
    ),
)
@click.option(
    '--config-yaml',
    '-y',
    type=click.Path(exists=True),
    default=None,
    help=(
        'Specify the path to a config YAML file with options for the worker. This is the same '
        'config block that you provide to dagster_celery.celery_executor when configuring a '
        'pipeline for execution with Celery, with, e.g., the URL of the broker to use.'
    ),
)
@click.option(
    '--queue',
    '-q',
    type=click.STRING,
    multiple=True,
    help=(
        'Names of the queues on which this worker should listen for tasks.  Provide multiple -q '
        'arguments to specify multiple queues. Note that each celery worker may listen on no more '
        'than four queues.'
    ),
)
@click.option(
    '--background', '-d', is_flag=True, help='Set this flag to run the worker in the background.'
)
@click.option(
    '--includes',
    '-i',
    type=click.STRING,
    multiple=True,
    help=(
        'Python modules the worker should import. Provide multiple -i arguments to specify '
        'multiple modules.'
    ),
)
@click.option(
    '--loglevel', '-l', type=click.STRING, default='INFO', help='Log level for the worker.'
)
def worker_start_command(
    name=None, config_yaml=None, background=None, queue=None, includes=None, loglevel=None,
):
    loglevel_args = ['--loglevel', loglevel]

    if len(queue) > 4:
        check.failed(
            'Can\'t start a dagster_celery worker that listens on more than four queues, due to a '
            'bug in Celery 4.'
        )
    queue_args = []
    for q in queue:
        queue_args += ['-Q', q]

    includes_args = ['-I', ','.join(includes)] if includes else []

    pythonpath = get_config_dir(config_yaml)
    subprocess_args = (
        ['celery', '-A', 'dagster_celery.tasks', 'worker', '--prefetch-multiplier=1']
        + loglevel_args
        + queue_args
        + includes_args
        + ['-n', get_worker_name(name)]
    )

    env = os.environ.copy()
    if pythonpath is not None:
        env['PYTHONPATH'] = '{existing_pythonpath}{pythonpath}:'.format(
            existing_pythonpath=env.get('PYTHONPATH', ''), pythonpath=pythonpath
        )

    if background:
        launch_background_worker(subprocess_args, env=env)
    else:
        return subprocess.check_call(subprocess_args, env=env)


@click.command(
    name='list',
    help='List running dagster-celery workers. Note that we use the broker to contact the workers.',
)
@click.option(
    '--config-yaml',
    '-y',
    type=click.Path(exists=True),
    default=None,
    help=(
        'Specify the path to a config YAML file with options for the workers you are trying to '
        'manage. This is the same config block that you provide to dagster_celery.celery_executor '
        'when configuring a pipeline for execution with Celery, with, e.g., the URL of the broker '
        'to use. Without this config file, you will not be able to find your workers (since the '
        'CLI won\'t know how to reach the broker).'
    ),
)
def worker_list_command(config_yaml=None):
    app = get_app(config_yaml)

    print(app.control.inspect(timeout=1).active())


@click.command(
    name='terminate',
    help=(
        'Shut down dagster-celery workers. Note that we use the broker to send signals to the '
        'workers to terminate -- if the broker is not running, this command is a no-op. '
        'Provide the argument NAME to terminate a specific worker by name.'
    ),
)
@click.argument('name', default='dagster')
@click.option(
    '--all', '-a', 'all_', is_flag=True, help='Set this flag to terminate all running workers.'
)
@click.option(
    '--config-yaml',
    '-y',
    type=click.Path(exists=True),
    default=None,
    help=(
        'Specify the path to a config YAML file with options for the workers you are trying to '
        'manage. This is the same config block that you provide to dagster_celery.celery_executor '
        'when configuring a pipeline for execution with Celery, with, e.g., the URL of the broker '
        'to use. Without this config file, you will not be able to terminate your workers (since '
        'the CLI won\'t know how to reach the broker).'
    ),
)
def worker_terminate_command(name='dagster', config_yaml=None, all_=False):
    app = get_app(config_yaml)

    if all_:
        app.control.broadcast('shutdown')
    else:
        app.control.broadcast(
            'shutdown', destination=[host_format(default_nodename(get_worker_name(name)))]
        )


worker_cli = create_worker_cli_group()


@click.group(commands={'worker': worker_cli})
def main():
    '''dagster-celery'''


if __name__ == '__main__':
    # pylint doesn't understand click
    main()  # pylint:disable=no-value-for-parameter
